from inspect import signature

from ..errors import Validation as errors


class Field:
  def __init__(self, context, instruction, parent, from_empty=False):
    self.context = context
    self.instruction = instruction
    self.name = instruction['name'] if 'name' in instruction else None
    self.parent = parent
    self._value = instruction['value'] if 'value' in instruction else None
    self.touched = False

    if from_empty:
      return

    instruction['element'] = self

    if instruction['type'] == 'BLOCK' and 'content_range' in instruction:
      self._value = context['input'][instruction['content_range'][0]:instruction['content_range'][1] + 1]

      for subinstruction in instruction['subinstructions']:
        subinstruction['element'] = self
    elif 'subinstructions' in instruction:
      unresolved_newlines = 0

      for subinstruction in instruction['subinstructions']:
        subinstruction['element'] = self

        if subinstruction['type'] == 'CONTINUATION':
          if self._value:
            if subinstruction['value']:
              if unresolved_newlines > 0:
                self._value += '\n' * unresolved_newlines
                unresolved_newlines = 0

              self._value += subinstruction['separator'] + subinstruction['value']
            elif subinstruction['separator'] == '\n':
              unresolved_newlines += 1
          else:
            self._value = subinstruction['value']
        elif subinstruction['type'] == 'BLOCK' and 'content_range' in subinstruction:
          self._value = context['input'][subinstruction['content_range'][0]:subinstruction['content_range'][1] + 1]

  def __repr__(self):
    value = self._value

    if value:
      value = self._value.replace("\n", '\n')
      if len(value) > 14:
        value = f"{value[0:11]}..."
      value = f"\"{value}\""
    else:
      value = 'None'

    if self.name:
      return f"<class Field name=\"{self.name}\" value={value}>"
    else:
      return f"<class Field value={value}>"

  def is_empty(self):
    return self._value is None

  def error(self, message=None):
    if callable(message):
      message = message(self.name, self._value)

    return errors.value_error(self.context, message, self.instruction)

  def raw(self):
    if self.name:
      return { self.name: self._value }
    else:
      return self._value

  def touch(self):
    self.touched = True

  def value(self, loader=None, *, enforce_value=False, required=None):
    if required is not None:
      enforce_value = required

    self.touched = True

    if self._value:
      if not loader:
        return self._value

      parameter_count = len(signature(loader).parameters)

      # TODO: Check if ValueError is 1) a valid pythonic approach 2) the best choice for the enopy API
      try:
        if parameter_count == 3:
          return loader(self.name, self._value, self.context)
        elif parameter_count == 1:
          return loader(self._value)
        else:
          return loader(self.name, self._value)
      except ValueError as message:
        raise errors.value_error(self.context, message, self.instruction)
    else:
      if enforce_value:
        raise errors.missing_value(self.context, self.instruction)

      return None
