from . import Empty, Field, Fieldset, List
from ..errors import Validation as errors


class Section:
  def __init__(self, context, instruction, parent):
    self.context = context
    self.depth = instruction['depth']
    self.instruction = instruction
    self.name = instruction['name']
    self.parent = parent

    self._elements = []
    self.elements_associative = {}
    self._enforce_all_elements = False
    self.touched = False

    instruction['element'] = self

    for subinstruction in instruction['subinstructions']:
      if subinstruction['type'] == 'EMPTY_LINE' or subinstruction['type'] == 'COMMENT':
        subinstruction['element'] = self
      elif subinstruction['type'] == 'NAME':
        self._append(Empty(context, subinstruction, self))
      elif subinstruction['type'] == 'FIELD':
        self._append(Field(context, subinstruction, self))
      elif subinstruction['type'] == 'LIST':
        self._append(List(context, subinstruction, self))
      elif subinstruction['type'] == 'BLOCK':
        self._append(Field(context, subinstruction, self))
      elif subinstruction['type'] == 'FIELDSET':
        self._append(Fieldset(context, subinstruction, self))
      elif subinstruction['type'] == 'SECTION':
        self._append(Section(context, subinstruction, self))

  def __repr__(self):
    if self.name == '<>#:=|\\_ENO_DOCUMENT':
      return f"<class Section document elements={len(self._elements)}>"
    else:
      return f"<class Section name=\"{self.name}\" elements={len(self._elements)}>"

  def _append(self, element):
    self._elements.append(element)

    if element.name in self.elements_associative:
      self.elements_associative[element.name].append(element)
    else:
      self.elements_associative[element.name] = [element]

  def assert_all_touched(self, message=None, *, only=None, skip=None):
    for name, elements in self.elements_associative.items():
      if (skip and name in skip) or (only and name not in only):
        continue

      for element in elements:
        if not element.touched:
          if callable(message):
            if (isinstance(element, Fieldset) or
                isinstance(element, Section)):
              value = None
            else:
              value = element.value()

            message = message(element.name, value)

          raise errors.excess_name(self.context, message, element.instruction)

        if isinstance(element, Fieldset) or isinstance(element, Section):
          element.assert_all_touched(message)

  def fieldsets(self, name):
    self.touched = True

    if name not in self.elements_associative:
      return []

    elements = self.elements_associative[name]

    for element in elements:
      if isinstance(element, Fieldset) or isinstance(element, Empty):
        element.touch()
      elif isinstance(element, List):
        raise errors.expected_fieldsets_got_list(self.context, element.instruction)
      elif isinstance(element, Section):
        raise errors.expected_fieldsets_got_section(self.context, element.instruction)
      elif isinstance(element, Field):
        raise errors.expected_fieldsets_got_field(self.context, element.instruction)

    # TODO: This is not very elegant
    return [(Fieldset(self.context, element.instruction, self, True) if isinstance(element, Empty) else element) for element in elements]

  def fieldset(self, name, *, enforce_element=True, required=None):
    if required is not None:
      enforce_element = required

    self.touched = True

    if name not in self.elements_associative:
      if enforce_element:
        raise errors.missing_fieldset(self.context, name, self.instruction)
      return None

    elements = self.elements_associative[name]

    for element in elements:
      if isinstance(element, Fieldset) or isinstance(element, Empty):
        continue
      elif isinstance(element, List):
        raise errors.expected_fieldset_got_list(self.context, element.instruction)
      elif isinstance(element, Section):
        raise errors.expected_fieldset_got_section(self.context, element.instruction)
      elif isinstance(element, Field):
        raise errors.expected_fieldset_got_field(self.context, element.instruction)

    if len(elements) > 1:
      raise errors.expected_fieldset_got_fieldsets(
        self.context,
        name,
        [e.instruction for e in elements]
      )

    element = elements[0]

    element.touch()

    if isinstance(element, Empty):
      return Fieldset(self.context, element.instruction, self, True)

    return element

  def element(self, name, *, enforce_element=True, required=None):
    if required is not None:
      enforce_element = required

    self.touched = True

    if name not in self.elements_associative:
      if enforce_element:
        raise errors.missing_element(self.context, name, self.instruction)

      return None

    elements = self.elements_associative[name]

    if len(elements) > 1:
      raise errors.expected_element_got_elements(self.context, name, [e.instruction for e in elements])

    element = elements[0]
    element.touch()

    return element

  def elements(self):
    self.touched = True

    for element in self._elements:
      element.touch()

    return self._elements

  def enforce_all_elements(self, enforce=True):
    self._enforce_all_elements = enforce

    for element in self._elements:
      if isinstance(element, Fieldset) or isinstance(element, Section):
        element.enforce_all_elements()

  def field(self,
            name,
            loader=None,
            *,
            enforce_element=None,
            enforce_value=False,
            required=None,
            with_element=False):

    if required is not None:
      enforce_value = required
    if enforce_element is None:
      enforce_element = self._enforce_all_elements
    enforce_element = enforce_element or enforce_value

    self.touched = True

    if name not in self.elements_associative:
      if enforce_element:
        raise errors.missing_field(self.context, name, self.instruction)

      return [None, None] if with_element else None

    elements = self.elements_associative[name]

    for element in elements:
      if isinstance(element, Field) or isinstance(element, Empty):
        continue
      elif isinstance(element, Fieldset):
        raise errors.expected_field_got_fieldset(self.context, element.instruction)
      elif isinstance(element, List):
        raise errors.expected_field_got_list(self.context, element.instruction)
      elif isinstance(element, Section):
        raise errors.expected_field_got_section(self.context, element.instruction)

    if len(elements) > 1:
      # TODO: Consider 'expected_[type]_got_multiple_elements' (*) for this here and also everywhere else,
      #       then this can be thrown as an accurate first error branch if there are multiple elements
      #       and the other type check errors are only triggered/checked if there is only a single element
      #       present, as requested indeed here.
      #
      #       (*) or even a more generic 'expected_single_got_multiple_elements' allowing more usecases additionally?
      raise errors.expected_field_got_fields(
        self.context,
        name,
        [e.instruction for e in elements]
      )

    element = elements[0]

    element.touch()

    if isinstance(element, Empty):
      element = Field(self.context, element.instruction, self, True)

    if with_element:
      return [element, element.value(loader, enforce_value=enforce_value)]

    return element.value(loader, enforce_value=enforce_value)

  def fields(self, name):
    self.touched = True

    if name not in self.elements_associative:
      return []

    elements = self.elements_associative[name]

    for element in elements:
      if isinstance(element, Field) or isinstance(element, Empty):
        element.touch()
      elif isinstance(element, Fieldset):
        raise errors.expected_fields_got_fieldset(self.context, element.instruction)
      elif isinstance(element, List):
        raise errors.expected_fields_got_list(self.context, element.instruction)
      elif isinstance(element, Section):
        raise errors.expected_fields_got_section(self.context, element.instruction)

    # TODO: This is not very elegant
    return [(Field(self.context, element.instruction, self, True) if isinstance(element, Empty) else element) for element in elements]

  def list(self,
           name,
           loader=None,
           *,
           enforce_element=None,
           enforce_values=True,
           exact_count=None,
           max_count=None,
           min_count=None,
           required=None,
           with_elements=False):

    if required is not None:
      enforce_element = required
    if enforce_element is None:
      enforce_element = self._enforce_all_elements
    enforce_element = enforce_element or exact_count and exact_count > 0 or min_count and min_count > 0

    self.touched = True

    if name not in self.elements_associative:
      if enforce_element:
        raise errors.missing_list(self.context, name, self.instruction)
      return []

    elements = self.elements_associative[name]

    for element in elements:
      if isinstance(element, List) or isinstance(element, Empty):
        continue
      elif isinstance(element, Field):
        raise errors.expected_list_got_field(self.context, element.instruction)
      elif isinstance(element, Fieldset):
        raise errors.expected_list_got_fieldset(self.context, element.instruction)
      elif isinstance(element, Section):
        raise errors.expected_list_got_section(self.context, element.instruction)

    if len(elements) > 1:
      raise errors.expected_list_got_lists(
        self.context,
        name,
        [e.instruction for e in elements]
      )

    element = elements[0]

    element.touch()

    count = 0 if isinstance(element, Empty) else element.length()

    if exact_count and count != exact_count:
      raise errors.exact_count_not_met(
        self.context,
        element.instruction,
        exact_count
      )

    if min_count and count < min_count:
      raise errors.min_count_not_met(
        self.context,
        element.instruction,
        min_count
      )

    if max_count and count > max_count:
      raise errors.max_count_not_met(
        self.context,
        element.instruction,
        max_count
      )

    if isinstance(element, Empty):
      return []
    else:
      return element.items(loader, enforce_values=enforce_values, with_elements=with_elements)

  def lists(self, name):
    self.touched = True

    if name not in self.elements_associative:
      return []

    elements = self.elements_associative[name]

    for element in elements:
      if isinstance(element, List) or isinstance(element, Empty):
        element.touch()
      if isinstance(element, Fieldset):
        raise errors.expected_lists_got_fieldset(self.context, element.instruction)
      elif isinstance(element, Section):
        raise errors.expected_lists_got_section(self.context, element.instruction)
      elif isinstance(element, Field):
        raise errors.expected_lists_got_field(self.context, element.instruction)

    # TODO: This is not very elegant
    return [(List(self.context, element.instruction, self, True) if isinstance(element, Empty) else element) for element in elements]

  def lookup(self, index_or_line, column=None):
    if column is not None:
      line = index_or_line
    else:
      index = 0
      line = 0
      column = 0

      while index != index_or_line:
        if index >= len(self.context['input']):
          return None

        char = self.context['input'][index]

        if char == '\n':
          line += 1
          column = 0
        else:
          column += 1

        index += 1

    instruction = next((i for i in self.context['instructions'] if i['line'] == line), None)

    if not instruction:
      return None

    result = {
      'element': instruction['element'],
      'zone': 'element'
    }

    if 'ranges' not in instruction:
      return result

    rightmost_match = 0
    for type, range in instruction['ranges'].items():
      if column >= range[0] and column <= range[1] and column >= rightmost_match:
        result['zone'] = type
        rightmost_match = column

    return result

  def raw(self):
    elements = [e.raw() for e in self._elements]

    if self.name == '<>#:=|\\_ENO_DOCUMENT':
      return elements

    return { self.name: elements }

  def section(self, name, *, enforce_element=True, required=None):
    if required is not None:
      enforce_element = required

    self.touched = True

    if name not in self.elements_associative:
      if enforce_element:
        raise errors.missing_section(self.context, name, self.instruction)
      return None

    elements = self.elements_associative[name]

    for element in elements:
      if isinstance(element, Section):
        continue
      elif isinstance(element, Fieldset):
        raise errors.expected_section_got_fieldset(self.context, element.instruction)
      elif isinstance(element, Empty):
        raise errors.expected_section_got_empty(self.context, element.instruction)
      elif isinstance(element, List):
        raise errors.expected_section_got_list(self.context, element.instruction)
      elif isinstance(element, Field):
        raise errors.expected_section_got_field(self.context, element.instruction)

    if len(elements) > 1:
      raise errors.expected_section_got_sections(self.context,
                                                 name,
                                                 [e.instruction for e in elements])

    elements[0].touch()

    return elements[0]

  def sections(self, name):
    self.touched = True

    if name not in self.elements_associative:
      return []

    elements = self.elements_associative[name]

    for element in elements:
      if isinstance(element, Section):
        element.touch()
      if isinstance(element, Fieldset):
        raise errors.expected_sections_got_fieldset(self.context, element.instruction)
      elif isinstance(element, Empty):
        raise errors.expected_sections_got_empty(self.context, element.instruction)
      elif isinstance(element, List):
        raise errors.expected_sections_got_list(self.context, element.instruction)
      elif isinstance(element, Field):
        raise errors.expected_sections_got_field(self.context, element.instruction)

    return elements

  def touch(self):
    self.touched = True
