from ..error_types import ValidationError


class Validation:
  @staticmethod
  def deep_expand_instruction(instruction):
    if 'subinstructions' not in instruction:
      return [instruction]

    results = [instruction]

    for subinstruction in instruction['subinstructions']:
      results.extend(Validation.deep_expand_instruction(subinstruction))

    return results

  @staticmethod
  def expand_instructions(instructions):
    results = []

    for instruction in instructions:
      results.append(instruction)
      if 'subinstructions' in instruction:
        results.extend(instruction['subinstructions'])

    return results

  @staticmethod
  def exact_count_not_met(context, instruction, expected_count):
    message = context['messages']['validation']['exact_count_not_met'](
      instruction['name'],
      len(instruction['subinstructions']) if 'subinstructions' in instruction else 0,
      expected_count
    )

    if 'subinstructions' in instruction and len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection = [
        [instruction['subinstructions'][0]['line'], 0],
        [last_subinstruction['line'], last_subinstruction['length']]
      ]
      snippet = context['reporter'].report(context, instruction['subinstructions'], instruction)
    else:
      selection = [
        [instruction['line'], instruction['length']],
        [instruction['line'], instruction['length']]
      ]
      snippet = context['reporter'].report(context, instruction)

    return ValidationError(message, snippet, selection)

  @staticmethod
  def excess_name(context, message, instruction):
    if message is None:
      message = context['messages']['validation']['excess_name'](instruction['name'])

    selection = [[instruction['line'], 0]]
    if 'subinstructions' in instruction and len(instruction['subinstructions']) > 0:
      snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      snippet = context['reporter'].report(context, instruction)
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_fieldsets_got_field(context, instruction):
    message = context['messages']['validation']['expected_fieldsets_got_field'](instruction['name'])
    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])
    selection = [[instruction['line'], 0]]

    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_fieldsets_got_list(context, instruction):
    message = context['messages']['validation']['expected_fieldsets_got_list'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_fieldsets_got_section(context, instruction):
    message = context['messages']['validation']['expected_fieldsets_got_section'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_fieldset_got_fieldsets(context, name, instructions):
    expanded_instructions = Validation.expand_instructions(instructions)

    message = context['messages']['validation']['expected_fieldset_got_fieldsets'](name)

    snippet = context['reporter'].report(context, instructions, expanded_instructions)

    selection = [
      [expanded_instructions[0]['line'], 0],
      [expanded_instructions[-1]['line'], expanded_instructions[-1]['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_fieldset_got_field(context, instruction):
    message = context['messages']['validation']['expected_fieldset_got_field'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_fieldset_got_list(context, instruction):
    message = context['messages']['validation']['expected_fieldset_got_list'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_fieldset_got_section(context, instruction):
    message = context['messages']['validation']['expected_fieldset_got_section'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_element_got_elements(context, name, instructions):
    expanded_instructions = Validation.expand_instructions(instructions)

    message = context['messages']['validation']['expected_element_got_elements'](name)

    snippet = context['reporter'].report(context, instructions, expanded_instructions)

    selection = [
      [expanded_instructions[0]['line'], 0],
      [expanded_instructions[-1]['line'], expanded_instructions[-1]['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_field_got_fieldset(context, instruction):
    message = context['messages']['validation']['expected_field_got_fieldset'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_field_got_fields(context, name, instructions):
    expanded_instructions = Validation.expand_instructions(instructions)

    message = context['messages']['validation']['expected_field_got_fields'](name)

    snippet = context['reporter'].report(context, instructions, expanded_instructions)

    selection = [
      [expanded_instructions[0]['line'], 0],
      [expanded_instructions[-1]['line'], expanded_instructions[-1]['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_field_got_list(context, instruction):
    message = context['messages']['validation']['expected_field_got_list'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_field_got_section(context, instruction):
    message = context['messages']['validation']['expected_field_got_section'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_fields_got_fieldset(context, instruction):
    message = context['messages']['validation']['expected_fields_got_fieldset'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_fields_got_list(context, instruction):
    message = context['messages']['validation']['expected_fields_got_list'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_fields_got_section(context, instruction):
    message = context['messages']['validation']['expected_fields_got_section'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_list_got_fieldset(context, instruction):
    message = context['messages']['validation']['expected_list_got_fieldset'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_list_got_field(context, instruction):
    message = context['messages']['validation']['expected_list_got_field'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_list_got_lists(context, name, instructions):
    expanded_instructions = Validation.expand_instructions(instructions)

    message = context['messages']['validation']['expected_list_got_lists'](name)

    snippet = context['reporter'].report(context, instructions, expanded_instructions)

    selection = [
      [expanded_instructions[0]['line'], 0],
      [expanded_instructions[-1]['line'], expanded_instructions[-1]['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_list_got_section(context, instruction):
    message = context['messages']['validation']['expected_list_got_section'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_lists_got_fieldset(context, instruction):
    message = context['messages']['validation']['expected_lists_got_fieldset'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_lists_got_field(context, instruction):
    message = context['messages']['validation']['expected_lists_got_field'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_lists_got_section(context, instruction):
    message = context['messages']['validation']['expected_lists_got_section'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_section_got_fieldset(context, instruction):
    message = context['messages']['validation']['expected_section_got_fieldset'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_section_got_empty(context, instruction):
    message = context['messages']['validation']['expected_section_got_empty'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, instruction)

    selection = [
      [instruction['line'], 0],
      [instruction['line'], instruction['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_section_got_field(context, instruction):
    message = context['messages']['validation']['expected_section_got_field'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_section_got_list(context, instruction):
    message = context['messages']['validation']['expected_section_got_list'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_section_got_sections(context, name, instructions):
    expanded_instructions = Validation.expand_instructions(instructions)

    message = context['messages']['validation']['expected_section_got_sections'](name)

    snippet = context['reporter'].report(context, instructions, expanded_instructions)

    selection = [
      [expanded_instructions[0]['line'], 0],
      [expanded_instructions[-1]['line'], expanded_instructions[-1]['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_sections_got_fieldset(context, instruction):
    message = context['messages']['validation']['expected_sections_got_fieldset'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_sections_got_empty(context, instruction):
    message = context['messages']['validation']['expected_sections_got_empty'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_sections_got_field(context, instruction):
    message = context['messages']['validation']['expected_sections_got_field'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def expected_sections_got_list(context, instruction):
    message = context['messages']['validation']['expected_sections_got_list'](
      instruction['name']
    )

    snippet = context['reporter'].report(context, [instruction, *instruction['subinstructions']])

    selection = [[instruction['line'], 0]]
    if len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def max_count_not_met(context, instruction, max_count):
    message = context['messages']['validation']['max_count_not_met'](
      instruction['name'],
      len(instruction['subinstructions']) if 'subinstructions' in instruction else 0,
      max_count
    )

    if 'subinstructions' in instruction and len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection = [
        [instruction['subinstructions'][0]['line'], 0],
        [last_subinstruction['line'], last_subinstruction['length']]
      ]
      snippet = context['reporter'].report(context, instruction['subinstructions'], instruction)
    else:
      selection = [
        [instruction['line'], instruction['length']],
        [instruction['line'], instruction['length']]
      ]
      snippet = context['reporter'].report(context, instruction)

    return ValidationError(message, snippet, selection)

  @staticmethod
  def min_count_not_met(context, instruction, min_count):
    message = context['messages']['validation']['min_count_not_met'](
      instruction['name'],
      len(instruction['subinstructions']) if 'subinstructions' in instruction else 0,
      min_count
    )

    if 'subinstructions' in instruction and len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection = [
        [instruction['subinstructions'][0]['line'], 0],
        [last_subinstruction['line'], last_subinstruction['length']]
      ]
      snippet = context['reporter'].report(context, instruction['subinstructions'], instruction)
    else:
      selection = [
        [instruction['line'], instruction['length']],
        [instruction['line'], instruction['length']]
      ]
      snippet = context['reporter'].report(context, instruction)

    return ValidationError(message, snippet, selection)

  @staticmethod
  def missing_fieldset(context, name, section_instruction):
    message = context['messages']['validation']['missing_fieldset'](name)

    snippet = context['reporter'].report(
      context,
      section_instruction,
      Validation.deep_expand_instruction(section_instruction)
    )

    selection = [
      [section_instruction['line'], section_instruction['length']],
      [section_instruction['line'], section_instruction['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def missing_fieldset_entry(context, name, fieldset_instruction):
    message = context['messages']['validation']['missing_fieldset_entry'](name)

    snippet = context['reporter'].report(
      context,
      fieldset_instruction,
      Validation.deep_expand_instruction(fieldset_instruction)
    )

    selection = [
      [fieldset_instruction['line'], fieldset_instruction['length']],
      [fieldset_instruction['line'], fieldset_instruction['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def missing_element(context, name, section_instruction):
    message = context['messages']['validation']['missing_element'](name)

    snippet = context['reporter'].report(
      context,
      section_instruction,
      Validation.deep_expand_instruction(section_instruction)
    )

    selection = [
      [section_instruction['line'], section_instruction['length']],
      [section_instruction['line'], section_instruction['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def missing_field(context, name, section_instruction):
    message = context['messages']['validation']['missing_field'](name)

    snippet = context['reporter'].report(
      context,
      section_instruction,
      Validation.deep_expand_instruction(section_instruction)
    )

    selection = [
      [section_instruction['line'], section_instruction['length']],
      [section_instruction['line'], section_instruction['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def missing_list(context, name, section_instruction):
    message = context['messages']['validation']['missing_list'](name)

    snippet = context['reporter'].report(
      context,
      section_instruction,
      Validation.deep_expand_instruction(section_instruction)
    )

    selection = [
      [section_instruction['line'], section_instruction['length']],
      [section_instruction['line'], section_instruction['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def missing_section(context, name, section_instruction):
    message = context['messages']['validation']['missing_section'](name)

    snippet = context['reporter'].report(
      context,
      section_instruction,
      Validation.deep_expand_instruction(section_instruction)
    )

    selection = [
      [section_instruction['line'], section_instruction['length']],
      [section_instruction['line'], section_instruction['length']]
    ]

    return ValidationError(message, snippet, selection)

  @staticmethod
  def missing_value(context, instruction):
    if instruction['type'] == 'FIELD' or instruction['type'] == 'NAME':
      message = context['messages']['validation']['missing_field_value'](instruction['name'])
    elif instruction['type'] == 'FIELDSET_ENTRY':
      message = context['messages']['validation']['missing_fieldset_entry_value'](instruction['name'])
    elif instruction['type'] == 'LIST_ITEM':
      message = context['messages']['validation']['missing_list_item_value'](instruction['name'])

    snippet = context['reporter'].report(
      context,
      instruction,
      Validation.deep_expand_instruction(instruction)
    )

    selection = []
    if 'template' in instruction['ranges']:
      selection.append([instruction['line'], instruction['ranges']['template'][1]])
    elif 'name_operator' in instruction['ranges']:
      selection.append([instruction['line'],
                       min(instruction['ranges']['name_operator'][1] + 1, instruction['length'])])
    elif 'entry_operator' in instruction['ranges']:
      selection.append([instruction['line'],
                       min(instruction['ranges']['entry_operator'][1] + 1, instruction['length'])])
    elif 'item_operator' in instruction['ranges']:
      selection.append([instruction['line'],
                       min(instruction['ranges']['item_operator'][1] + 1, instruction['length'])])

    if 'subinstructions' in instruction and len(instruction['subinstructions']) > 0:
      last_subinstruction = instruction['subinstructions'][-1]
      selection.append([last_subinstruction['line'], last_subinstruction['length']])
    else:
      selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)

  @staticmethod
  def value_error(context, message, instruction):
    message = message or context['messages']['validation']['generic_error'](instruction['name'])

    if instruction['type'] == 'BLOCK':
      content_instructions = [i for i in instruction['subinstructions'] if i['type'] == 'BLOCK_CONTENT']
      terminator_instruction = instruction['subinstructions'][-1]

      if len(content_instructions) > 0:
        snippet = context['reporter'].report(context, content_instructions)
        selection = [
          [content_instructions[0]['line'], content_instructions[0]['ranges']['content'][0]],
          [content_instructions[-1]['line'], content_instructions[-1]['ranges']['content'][1]]
        ]
      else:
        snippet = context['reporter'].report(context, [instruction, terminator_instruction])
        selection = [
          [instruction['line'], instruction['length']],
          [instruction['line'], instruction['length']]
        ]
    else:
      snippet = context['reporter'].report(context, Validation.deep_expand_instruction(instruction))

      if 'value' in instruction['ranges']:
        selection = [[instruction['line'], instruction['ranges']['value'][0]]]
      elif 'template' in instruction['ranges']:
        selection = [[instruction['line'], instruction['ranges']['template_operator'][0]]]
      elif 'name_operator' in instruction['ranges']:
        selection = [[
          instruction['line'],
          min(instruction['ranges']['name_operator'][1] + 1, instruction['length']) # TODO: Is there a discrepancy between range indexing (+1 on the [1] token) and instruction length (true indices?)
        ]]
      elif 'entry_operator' in instruction['ranges']:
        selection = [[
          instruction['line'],
          min(instruction['ranges']['entry_operator'][1] + 1, instruction['length'])
        ]]
      elif instruction['type'] == 'LIST_ITEM':
        selection = [[
          instruction['line'],
          min(instruction['ranges']['item_operator'][1] + 1, instruction['length'])
        ]]
      else:
        selection = [[instruction['line'], instruction['length']]]

      if instruction['subinstructions'] and len(instruction['subinstructions']) > 0:
        last_instruction = instruction['subinstructions'][-1]
        selection.append([last_instruction['line'], last_instruction['length']])
      else:
        if 'value' in instruction['ranges']:
          selection.append([instruction['line'], instruction['ranges']['value'][1]])
        else:
          selection.append([instruction['line'], instruction['length']])

    return ValidationError(message, snippet, selection)
