from datetime import date as python_date, datetime as python_datetime
import json as python_json
import re

python_float = float

COLOR_RE = re.compile(r'^\s*#[0-9a-f]{3}([0-9a-f]{3})?\s*$', re.IGNORECASE)
DATE_RE = re.compile(r'^\s*(\d{4})-(\d\d)-(\d\d)\s*$')
DATETIME_RE = re.compile(r'^\s*(\d{4})(?:-(\d\d)(?:-(\d\d)(?:T(\d\d):(\d\d)(?::(\d\d)(?:\.(\d+))?)?(?:(Z)|([+\-])(\d\d):(\d\d)))?)?)?\s*$')
EMAIL_RE = re.compile(r'^\s*[^@\s]+@[^@\s]+\.[^@\s]+\s*$')
FLOAT_RE = re.compile(r'^\s*-?\d+(\.\d+)?\s*$')
INTEGER_RE = re.compile(r'^\s*-?\d+\s*$')
LAT_LNG_RE = re.compile(r'^\s*(-?\d{1,3}(?:\.\d+)?)\s*,\s*(-?\d{1,3}(?:\.\d+)?)\s*$')
URL_RE = re.compile(r'^\s*https?:\/\/[^\s.]+\.\S+\s*$')

def boolean(name, value, context):
  lower = value.strip().lower()

  if lower == 'true':
    return True
  if lower == 'false':
    return False
  if lower == 'yes':
    return True
  if lower == 'no':
    return False

  raise ValueError(context['messages']['loaders']['invalid_boolean'](name))

def color(name, value, context):
  if not COLOR_RE.match(value):
    raise ValueError(context['messages']['loaders']['invalid_color'](name))

  return value

def comma_separated(name, value, context):
  return [item.strip() for item in value.split(',')]

def date(name, value, context):
  match = DATE_RE.match(value)

  if not match:
    raise ValueError(context['messages']['loaders']['invalid_date'](name))

  year = int(match.group(1))
  month = int(match.group(2))
  day = int(match.group(3))

  return python_date(year, month, day)


# Format specification thankfully taken from https://www.w3.org/TR/NOTE-datetime
#
# 1997
# 1997-07
# 1997-07-16
# 1997-07-16T19:20+01:00
# 1997-07-16T19:20:30+01:00
# 1997-07-16T19:20:30.45+01:00
# 1994-11-05T08:15:30-05:00
# 1994-11-05T13:15:30Z
#
def datetime(name, value, context):
  match = DATETIME_RE.match(value)

  if not match:
    raise ValueError(context['messages']['loaders']['invalid_datetime'](name))

  year = match.group(1)
  month = match.group(2) or '01'
  day = match.group(3) or '01'
  hour = match.group(4) or '00'
  minute = match.group(5) or '00'
  second = match.group(6) or '00'
  fraction = match.group(7) or '00'
  zulu = match.group(8)

  if zulu is None:
    offset_sign = match.group(9) or '+'
    offset_hours = match.group(10) or '00'
    offset_minutes = match.group(11) or '00'
  else:
    offset_sign = '+'
    offset_hours = '00'
    offset_minutes = '00'

  normalized = f"{year}-{month}-{day}T{hour}:{minute}:{second}.{fraction}{offset_sign}{offset_hours}{offset_minutes}"

  # TODO: See if this can be replaced with a direct initialization without another string parse step at some point
  return python_datetime.strptime(normalized, '%Y-%m-%dT%H:%M:%S.%f%z')

def email(name, value, context):
  if not EMAIL_RE.match(value):
    raise ValueError(context['messages']['loaders']['invalid_email'](name))

  return value

def float(name, value, context):
  if not FLOAT_RE.match(value):
    raise ValueError(context['messages']['loaders']['invalid_float'](name))

  return python_float(value)

def integer(name, value, context):
  if not INTEGER_RE.match(value):
    raise ValueError(context['messages']['loaders']['invalid_integer'](name))

  return int(value)

def json(name, value, context):
  try:
    return python_json.loads(value)
  except python_json.JSONDecodeError as e:
    raise ValueError(context['messages']['loaders']['invalid_json'](name, e))

def lat_lng(name, value, context):
  match = LAT_LNG_RE.match(value)

  if not match:
    raise ValueError(context['messages']['loaders']['invalid_lat_lng'](name))

  return { 'lat': python_float(match.group(1)), 'lng': python_float(match.group(2)) }

def url(name, value, context):
  if not URL_RE.match(value):
    raise ValueError(context['messages']['loaders']['invalid_url'](name))

  return value

bootstrap_loaders = {
  'boolean': boolean,
  'color': color,
  'comma_separated': comma_separated,
  'date': date,
  'datetime': datetime,
  'email': email,
  'float': float,
  'integer': integer,
  'json': json,
  'lat_lng': lat_lng,
  'number': integer,
  'url': url
}
