from .errors import Resolution as errors


class Resolver:
  def __init__(self, context):
    self.context = context

  def resolve(self):
    while len(self.context['unresolved_instructions']) > 0:
      self.resolve_instruction(self.context['unresolved_instructions'][0])

  def consolidate(self, instruction, template):
    if instruction['type'] == 'NAME':
      if template['type'] == 'BLOCK':
        instruction['type'] = 'FIELD'
        self.copy_block(instruction, template)
      elif template['type'] == 'FIELD':
        instruction['type'] = 'FIELD'
        self.copy_field(instruction, template)
      elif template['type'] == 'FIELDSET':
        instruction['type'] = 'FIELDSET'
        self.copy_generic(instruction, template)
      elif template['type'] == 'LIST':
        instruction['type'] = 'LIST'
        self.copy_generic(instruction, template)
      elif template['type'] == 'SECTION':
        raise errors.copying_section_into_empty(self.context, instruction)
    elif instruction['type'] == 'SECTION':
      if template['type'] == 'SECTION':
        self.merge_sections(instruction, template, 'deep_copy' in instruction)
      elif template['type'] == 'BLOCK':
        raise errors.copying_block_into_section(self.context, instruction)
      elif template['type'] == 'FIELDSET':
        raise errors.copying_fieldset_into_section(self.context, instruction)
      elif template['type'] == 'FIELD':
        raise errors.copying_field_into_section(self.context, instruction)
      elif template['type'] == 'LIST':
        raise errors.copying_list_into_section(self.context, instruction)
    elif instruction['type'] == 'FIELDSET':
      if template['type'] == 'FIELDSET':
        self.merge_fieldsets(instruction, template)
      elif template['type'] == 'BLOCK':
        raise errors.copying_block_into_fieldset(self.context, instruction)
      elif template['type'] == 'FIELD':
        raise errors.copying_field_into_fieldset(self.context, instruction)
      elif template['type'] == 'LIST':
        raise errors.copying_list_into_fieldset(self.context, instruction)
      elif template['type'] == 'SECTION':
        raise errors.copying_section_into_fieldset(self.context, instruction)
    elif instruction['type'] == 'LIST':
      if template['type'] == 'LIST':
        self.copy_generic(instruction, template)
      elif template['type'] == 'BLOCK':
        raise errors.copying_block_into_list(self.context, instruction)
      elif template['type'] == 'FIELD':
        raise errors.copying_field_into_list(self.context, instruction)
      elif template['type'] == 'FIELDSET':
        raise errors.copying_fieldset_into_list(self.context, instruction)
      elif template['type'] == 'SECTION':
        raise errors.copying_section_into_list(self.context, instruction)
    elif instruction['type'] == 'FIELD':
      if template['type'] == 'FIELD':
        self.copy_field(instruction, template)
      elif template['type'] == 'BLOCK':
        self.copy_block(instruction, template)
      elif template['type'] == 'FIELDSET':
        raise errors.copying_fieldset_into_field(self.context, instruction)
      elif template['type'] == 'LIST':
        raise errors.copying_list_into_field(self.context, instruction)
      elif template['type'] == 'SECTION':
        raise errors.copying_section_into_field(self.context, instruction)

  def copy_block(self, instruction, template):
    instruction['subinstructions'].appendleft(template.copy())

  def copy_field(self, instruction, template):
    if template['value']:
      # TODO: Refactor this to make it more robust and less hack-ish
      instruction['subinstructions'].appendleft({
        'line': template['line'],
        'ranges': template['ranges'],
        'separator': '\n',
        'type': 'CONTINUATION',
        'value': template['value']
      })

    self.copy_generic(instruction, template)

  def copy_generic(self, instruction, template):
    for template_subinstruction in reversed(template['subinstructions']):
      if (template_subinstruction['type'] == 'EMPTY_LINE' or
          template_subinstruction['type'] == 'COMMENT'):
        continue

      instruction['subinstructions'].appendleft(template_subinstruction.copy())

  def merge_fieldsets(self, instruction, template):
    existing_entry_names = [i['name'] for i in instruction['subinstructions'] if i['type'] == 'FIELDSET_ENTRY']

    for template_subinstruction in reversed(template['subinstructions']):
      if template_subinstruction['type'] != 'FIELDSET_ENTRY':
        continue

      if template_subinstruction['name'] not in existing_entry_names:
        instruction['subinstructions'].appendleft(template_subinstruction.copy())

  def merge_sections(self, instruction, template, deep_merge):
    existing_subinstructions_name_index = {}

    for template_subinstruction in reversed(template['subinstructions']):
      if (template_subinstruction['type'] == 'EMPTY_LINE' or
          template_subinstruction['type'] == 'COMMENT'):
        continue

      if template_subinstruction['name'] not in existing_subinstructions_name_index:
        existing_subinstructions_name_index[template_subinstruction['name']] = [i for i in instruction['subinstructions'] if 'name' in i and i['name'] == template_subinstruction['name']]

      existing_subinstructions = existing_subinstructions_name_index[template_subinstruction['name']]

      if len(existing_subinstructions) == 0:
        instruction['subinstructions'].appendleft(template_subinstruction.copy())
        continue

      if not deep_merge or len(existing_subinstructions) > 1:
        continue

      if (template_subinstruction['type'] == 'FIELDSET' and
          existing_subinstructions[0]['type'] == 'FIELDSET'):

        template_subinstructions_with_same_name = [i for i in template['subinstructions'] if 'name' in i and i['name'] == template_subinstruction['name']]

        if len(template_subinstructions_with_same_name) == 1:
          self.merge_fieldsets(existing_subinstructions[0], template_subinstruction)

      if (template_subinstruction['type'] != 'SECTION' or
          existing_subinstructions[0]['type'] != 'SECTION'):
        continue

      template_subinstructions_with_same_name = [i for i in template['subinstructions'] if 'name' in i and i['name'] == template_subinstruction['name']]

      if len(template_subinstructions_with_same_name) == 1:
        self.merge_sections(existing_subinstructions[0], template_subinstruction, True)

  def resolve_instruction(self, instruction, previous_instructions=[]):
    if instruction['type'] == 'EMPTY_LINE' or instruction['type'] == 'COMMENT':
      return

    if instruction in previous_instructions:
      raise errors.cyclic_dependency(self.context, instruction, previous_instructions)

    if instruction['type'] == 'SECTION':
      for subinstruction in instruction['subinstructions']:
        self.resolve_instruction(subinstruction, [*previous_instructions, instruction])

    if 'template' not in instruction or 'consolidated' in instruction:
      return

    if instruction['template'] not in self.context['template_index']:
      raise errors.template_not_found(self.context, instruction)

    templates = self.context['template_index'][instruction['template']]

    if len(templates) > 1:
      raise errors.multiple_templates_found(self.context, instruction, templates)

    template = templates[0]

    self.resolve_instruction(template, [*previous_instructions, instruction])
    self.consolidate(instruction, template)

    instruction['consolidated'] = True

    self.context['unresolved_instructions'].remove(instruction)
