import re
from .errors import Tokenization as errors
from .grammar_regex import Grammar


class Tokenizer:
  def __init__(self, context):
    self.context = context
    self.input = context['input']
    self.line = 0
    self.index = 0
    self.instructions = []

    context['instructions'] = self.instructions

  def tokenize(self):
    while True:
      match = Grammar.REGEX.search(self.input, self.index)

      if not match or match.start() != self.index:
        instruction = self.tokenize_error_context()
        raise errors.invalid_line(self.context, instruction)

      instruction = {
        'index': self.index,
        'line': self.line
      }

      block = False

      if match.group(Grammar.EMPTY_LINE_INDEX) is not None:

        instruction['type'] = 'EMPTY_LINE'

      elif match.group(Grammar.NAME_OPERATOR_INDEX) is not None:

        unescaped_name = match.group(Grammar.NAME_UNESCAPED_INDEX)

        if unescaped_name is not None:
          instruction['name'] = unescaped_name

          name_column = match.start(Grammar.NAME_UNESCAPED_INDEX) - self.index
          name_operator_column = match.start(Grammar.NAME_OPERATOR_INDEX) - self.index

          instruction['ranges'] = {
            'name': [name_column, name_column + len(unescaped_name)],
            'name_operator': [name_operator_column, name_operator_column + 1]
          }
        else:
          escaped_name = match.group(Grammar.NAME_ESCAPED_INDEX)

          instruction['name'] = escaped_name

          escape_operator = match.group(Grammar.NAME_ESCAPE_BEGIN_OPERATOR_INDEX)
          escape_begin_operator_column = match.start(Grammar.NAME_ESCAPE_BEGIN_OPERATOR_INDEX) - self.index
          name_column = match.start(Grammar.NAME_ESCAPED_INDEX) - self.index
          escape_end_operator_column = match.start(Grammar.NAME_ESCAPE_END_OPERATOR_INDEX) - self.index
          name_operator_column = match.start(Grammar.NAME_OPERATOR_INDEX) - self.index

          instruction['ranges'] = {
            'escape_begin_operator': [escape_begin_operator_column, escape_begin_operator_column + len(escape_operator)],
            'escape_end_operator': [escape_end_operator_column, escape_end_operator_column + len(escape_operator)],
            'name_operator': [name_operator_column, name_operator_column + 1],
            'name': [name_column, name_column + len(escaped_name)]
          }

        value = match.group(Grammar.FIELD_VALUE_INDEX)
        if value is not None:
          instruction['type'] = 'FIELD'
          instruction['value'] = value

          value_column = match.start(Grammar.FIELD_VALUE_INDEX) - self.index
          instruction['ranges']['value'] = [value_column, value_column + len(value)]
        else:
          instruction['type'] = 'NAME'

      elif match.group(Grammar.LIST_ITEM_OPERATOR_INDEX) is not None:

        instruction['type'] = 'LIST_ITEM'

        operator_column = match.start(Grammar.LIST_ITEM_OPERATOR_INDEX) - self.index
        instruction['ranges'] = { 'item_operator': [operator_column, operator_column + 1] }

        value = match.group(Grammar.LIST_ITEM_VALUE_INDEX)
        instruction['value'] = value

        if value is not None:
          value_column = match.start(Grammar.LIST_ITEM_VALUE_INDEX) - self.index
          instruction['ranges']['value'] = [value_column, value_column + len(value)]

      elif match.group(Grammar.FIELDSET_ENTRY_OPERATOR_INDEX):

        unescaped_name = match.group(Grammar.NAME_UNESCAPED_INDEX)

        if unescaped_name is not None:
          instruction['name'] = unescaped_name

          name_column = match.start(Grammar.NAME_UNESCAPED_INDEX) - self.index
          entry_operator_column = match.start(Grammar.FIELDSET_ENTRY_OPERATOR_INDEX) - self.index

          instruction['ranges'] = {
            'entry_operator': [entry_operator_column, entry_operator_column + 1],
            'name': [name_column, name_column + len(unescaped_name)]
          }
        else:
          escaped_name = match.group(Grammar.NAME_ESCAPED_INDEX)

          instruction['name'] = escaped_name

          escape_operator = match.group(Grammar.NAME_ESCAPE_BEGIN_OPERATOR_INDEX)
          escape_begin_operator_column = match.start(Grammar.NAME_ESCAPE_BEGIN_OPERATOR_INDEX) - self.index
          name_column = match.start(Grammar.NAME_ESCAPED_INDEX) - self.index
          escape_end_operator_column = match.start(Grammar.NAME_ESCAPE_END_OPERATOR_INDEX) - self.index
          entry_operator_column = match.start(Grammar.FIELDSET_ENTRY_OPERATOR_INDEX) - self.index

          instruction['ranges'] = {
            'escape_begin_operator': [escape_begin_operator_column, escape_begin_operator_column + len(escape_operator)],
            'escape_end_operator': [escape_end_operator_column, escape_end_operator_column + len(escape_operator)],
            'entry_operator': [entry_operator_column, entry_operator_column + 1],
            'name': [name_column, name_column + len(escaped_name)]
          }

        instruction['type'] = 'FIELDSET_ENTRY'

        value = match.group(Grammar.FIELDSET_ENTRY_VALUE_INDEX)
        instruction['value'] = value

        if value is not None:
          value_column = match.start(Grammar.FIELDSET_ENTRY_VALUE_INDEX) - self.index
          instruction['ranges']['value'] = [value_column, value_column + len(value)]


      elif match.group(Grammar.LINE_CONTINUATION_OPERATOR_INDEX) is not None:

        instruction['separator'] = ' '
        instruction['type'] = 'CONTINUATION'

        value = match.group(Grammar.LINE_CONTINUATION_VALUE_INDEX)
        instruction['value'] = value

        operator_column = match.start(Grammar.LINE_CONTINUATION_OPERATOR_INDEX) - self.index
        instruction['ranges'] = { 'line_continuation_operator': [operator_column, operator_column + 1] }

        if value is not None:
          value_column = match.start(Grammar.LINE_CONTINUATION_VALUE_INDEX) - self.index
          instruction['ranges']['value'] = [value_column, value_column + len(value)]

      elif match.group(Grammar.NEWLINE_CONTINUATION_OPERATOR_INDEX) is not None:

        instruction['separator'] = '\n'
        instruction['type'] = 'CONTINUATION'

        value = match.group(Grammar.NEWLINE_CONTINUATION_VALUE_INDEX)
        instruction['value'] = value

        operator_column = match.start(Grammar.NEWLINE_CONTINUATION_OPERATOR_INDEX) - self.index
        instruction['ranges'] = { 'newline_continuation_operator': [operator_column, operator_column + 1] }

        if value is not None:
          value_column = match.start(Grammar.NEWLINE_CONTINUATION_VALUE_INDEX) - self.index
          instruction['ranges']['value'] = [value_column, value_column + len(value)]

      elif match.group(Grammar.SECTION_HASHES_INDEX) is not None:

        section_operator = match.group(Grammar.SECTION_HASHES_INDEX)

        instruction['depth'] = len(section_operator)
        instruction['type'] = 'SECTION'

        section_operator_column = match.start(Grammar.SECTION_HASHES_INDEX) - self.index
        unescaped_name = match.group(Grammar.SECTION_NAME_UNESCAPED_INDEX)

        if unescaped_name is not None:
          instruction['name'] = unescaped_name

          name_column = match.start(Grammar.SECTION_NAME_UNESCAPED_INDEX) - self.index
          name_end_column = name_column + len(unescaped_name)

          instruction['ranges'] = {
            'name': [name_column, name_end_column],
            'section_operator': [section_operator_column, section_operator_column + len(section_operator)]
          }
        else:
          escaped_name = match.group(Grammar.SECTION_NAME_ESCAPED_INDEX)
          instruction['name'] = escaped_name

          escape_operator = match.group(Grammar.SECTION_NAME_ESCAPE_BEGIN_OPERATOR_INDEX)
          escape_begin_operator_column = match.start(Grammar.SECTION_NAME_ESCAPE_BEGIN_OPERATOR_INDEX) - self.index
          name_column = match.start(Grammar.SECTION_NAME_ESCAPED_INDEX) - self.index
          escape_end_operator_column = match.start(Grammar.SECTION_NAME_ESCAPE_END_OPERATOR_INDEX) - self.index
          name_end_column = escape_end_operator_column + len(escape_operator)

          instruction['ranges'] = {
            'escape_begin_operator': [escape_begin_operator_column, escape_begin_operator_column + len(escape_operator)],
            'escape_end_operator': [escape_end_operator_column, escape_end_operator_column + len(escape_operator)],
            'name': [name_column, name_column + len(escaped_name)],
            'section_operator': [section_operator_column, section_operator_column + len(section_operator)]
          }

        template = match.group(Grammar.SECTION_TEMPLATE_INDEX)
        if template is not None:
          instruction['template'] = template

          copy_operator = match.group(Grammar.SECTION_COPY_OPERATOR_INDEX)
          copy_operator_column = match.start(Grammar.SECTION_COPY_OPERATOR_INDEX) - self.index
          template_column = match.start(Grammar.SECTION_TEMPLATE_INDEX) - self.index

          if copy_operator == '<':
            instruction['deep_copy'] = False
            instruction['ranges']['copy_operator'] = [copy_operator_column, copy_operator_column + 1]
          else: # copy_operator == '<<':
            instruction['deep_copy'] = True
            instruction['ranges']['deep_copy_operator'] = [copy_operator_column, copy_operator_column + 2]

          instruction['ranges']['template'] = [template_column, template_column + len(template)]

      elif match.group(Grammar.BLOCK_DASHES_INDEX) is not None:

        operator = match.group(Grammar.BLOCK_DASHES_INDEX)
        name = match.group(Grammar.BLOCK_NAME_INDEX)

        instruction['name'] = name
        instruction['type'] = 'BLOCK'

        operator_column = match.start(Grammar.BLOCK_DASHES_INDEX) - self.index
        name_column = match.start(Grammar.BLOCK_NAME_INDEX) - self.index
        instruction['length'] = match.end() - self.index
        instruction['ranges'] = {
          'block_operator': [operator_column, operator_column + len(operator)],
          'name': [name_column, name_column + len(name)]
        }

        self.context['instructions'].append(instruction)

        self.index = match.end()

        terminator_re = re.compile(f"\\n[^\\S\\n]*({operator})[^\\S\\n]*({re.escape(name)})[^\\S\\n]*(?=\\n|$)")
        terminator_match = terminator_re.search(self.input, self.index)

        self.index += 1  # move past current char (\n) into next line
        self.line += 1

        if not terminator_match:
          self.tokenize_error_context()
          raise errors.unterminated_block(self.context, instruction)

        end_of_block_index = terminator_match.start()

        if end_of_block_index != self.index - 1:
          instruction['content_range'] = [self.index, end_of_block_index - 1]

          while True:
            end_of_line_index = self.input.find('\n', self.index, end_of_block_index)

            if end_of_line_index == -1:
              self.instructions.append({
                'index': self.index,
                'length': end_of_block_index - self.index,
                'line': self.line,
                'ranges': { 'content': [0, end_of_block_index - self.index] },
                'type': 'BLOCK_CONTENT',
              })

              self.index = end_of_block_index + 1
              self.line += 1

              break
            else:
              self.instructions.append({
                'index': self.index,
                'length': end_of_line_index - self.index,
                'line': self.line,
                'ranges': { 'content': [0, end_of_line_index - self.index] },
                'type': 'BLOCK_CONTENT'
              })

              self.index = end_of_line_index + 1
              self.line += 1

        instruction = {
          'index': self.index,
          'length': terminator_match.end() - self.index,
          'line': self.line,
          'ranges': {
            'block_operator': [terminator_match.start(1) - self.index, terminator_match.end(1) - self.index],
            'name': [terminator_match.start(2) - self.index, terminator_match.end(2) - self.index]
          },
          'type': 'BLOCK_TERMINATOR'
        }

        self.instructions.append(instruction)

        self.index = terminator_match.end() + 1
        self.line += 1

        block = True

      elif match.group(Grammar.COMMENT_OPERATOR_INDEX) is not None:

        comment = match.group(Grammar.COMMENT_TEXT_INDEX)
        comment_operator_column = match.start(Grammar.COMMENT_OPERATOR_INDEX) - self.index

        instruction['type'] = 'COMMENT'
        instruction['comment'] = comment
        instruction['ranges'] = {
          'comment_operator': [comment_operator_column, comment_operator_column + 1],

        }

        if comment is not None:
          comment_column = match.start(Grammar.COMMENT_TEXT_INDEX) - self.index
          instruction['ranges']['comment'] = [comment_column, comment_column + len(comment)]

      elif match.group(Grammar.COPY_OPERATOR_INDEX) is not None:

        operator = match.group(Grammar.COPY_OPERATOR_INDEX)
        template = match.group(Grammar.TEMPLATE_INDEX)
        unescaped_name = match.group(Grammar.NAME_UNESCAPED_INDEX)

        if unescaped_name is not None:
          instruction['name'] = unescaped_name

          name_column = match.start(Grammar.NAME_UNESCAPED_INDEX) - self.index
          operator_column = match.start(Grammar.COPY_OPERATOR_INDEX) - self.index

          instruction['ranges'] = {
            'copy_operator': [operator_column, operator_column + len(operator)],
            'name': [name_column, name_column + len(unescaped_name)]
          }
        else:
          escaped_name = match.group(Grammar.NAME_ESCAPED_INDEX)
          instruction['name'] = escaped_name

          escape_operator = match.group(Grammar.NAME_ESCAPE_BEGIN_OPERATOR_INDEX)
          escape_begin_operator_column = match.start(Grammar.NAME_ESCAPE_BEGIN_OPERATOR_INDEX) - self.index
          name_column = match.start(Grammar.NAME_ESCAPED_INDEX) - self.index
          escape_end_operator_column = match.start(Grammar.NAME_ESCAPE_END_OPERATOR_INDEX) - self.index
          operator_column = match.start(Grammar.COPY_OPERATOR_INDEX) - self.index

          instruction['ranges'] = {
            'copy_operator': [operator_column, operator_column + len(operator)],
            'escape_begin_operator': [escape_begin_operator_column, escape_begin_operator_column + len(escape_operator)],
            'escape_end_operator': [escape_end_operator_column, escape_end_operator_column + len(escape_operator)],
            'name': [name_column, name_column + len(escaped_name)]
          }

        instruction['template'] = template
        instruction['type'] = 'NAME'

        template_column = match.start(Grammar.TEMPLATE_INDEX) - self.index
        instruction['ranges']['template'] = [template_column, template_column + len(template)]


      if not block:
        instruction['length'] = match.end() - self.index
        self.index = match.end() + 1
        self.context['instructions'].append(instruction)
        self.line += 1

      if self.index >= len(self.input):
        if len(self.input) > 0 and self.input[-1] == '\n':
          self.context['instructions'].append({
            'index': len(self.input),
            'length': 0,
            'line': self.line,
            'type': 'EMPTY_LINE'
          })

        break

  def tokenize_error_context(self):
    first_instruction = None

    while True:
      end_of_line_index = self.input.find('\n', self.index)

      if end_of_line_index == -1:
        instruction = {
          'index': self.index,
          'length': len(self.input) - self.index,
          'line': self.line
        }

        self.context['instructions'].append(instruction)

        if first_instruction is None:
          return instruction
        else:
          return first_instruction
      else:
        instruction = {
          'index': self.index,
          'length': end_of_line_index - self.index,
          'line': self.line
        }

        self.context['instructions'].append(instruction)

        if first_instruction is None:
          first_instruction = instruction

        self.index = end_of_line_index + 1
        self.line += 1
